/*
 * Downloaded from lenkaspace.net
 * 
 * Paper about this project: http://lenkaspace.net/papers/iteratedPrisonersDilemmaNeuralNetworks.php
 * Email: contact@lenkaspace.net
 */

package net.lenkaspace.adaptiveNavigation.model;

import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.util.ArrayList;

import net.lenkaspace.adaptiveNavigation.ControlsPanel;
import net.lenkaspace.adaptiveNavigation.Maths;
import net.lenkaspace.adaptiveNavigation.Settings;
import net.lenkaspace.adaptiveNavigation.reports.ReportController;
import net.lenkaspace.adaptiveNavigation.vo.Vector3D;


/** 
 * Agent
 * 	- corresponds to a robot, moves around using motors
 */
public class Agent extends BaseSituatedModel  {
	public static final int NUM_CELLS = 9;
	
	public static final int MOTOR_LEFT = 0;
	public static final int MOTOR_RIGHT = 1;
	public static final int MOTOR_BOTTOM = 2;
	public static final int MOTOR_TOP = 3;
	
	public static final int MAX_MOTOR_FORCE = 1;
	
	public static final int BODY_W = 62;
	public static final int BODY_H = 62;
	
	public static final double MAX_FOODLEVEL = 10.0;
	
	private ArrayList <Cell> cells;
	private double[] motorForces;
	
	private double foodLevel;
	private Vector3D oldWorldPosition;
	
	
	public Agent(int id_,Vector3D pos_, World world_)  {
		super(id_, pos_, new Vector3D (Agent.BODY_W,Agent.BODY_H,0), world_);
		motorForces = new double[4];	
		this.createCells();	
		this.reset();
	}
	
	/**
	 * Copies attributes from another agent, used during cloning
	 * @param agent
	 */
	public void copyAgent(Agent agent) {
		this.foodLevel = agent.foodLevel;
		for (int i=0; i<NUM_CELLS; i++) {
			Cell cell = this.cells.get(i);
			cell.copyFromCell(agent.getCells().get(i));
		}
	}
	
	/**
	 * Resets state variables and cells
	 */
	public void reset() {
		this.foodLevel = Agent.MAX_FOODLEVEL;
		this.position = new Vector3D (300,300,0);
		this.steeringForce = new Vector3D(0,0,0);
		for (Cell cell : this.cells) {
			cell.reset();
		}
		motorForces[Agent.MOTOR_LEFT] = 0;//0.5;
		motorForces[Agent.MOTOR_RIGHT] = 0;//1;
		this.oldWorldPosition = new Vector3D(this.getWorldPosition());
		this.initCellsPositions();
	}
	
	/**
	 * create and initialise cells
	 */
	public void createCells() {
		//create cells
		cells = new ArrayList <Cell> ();
		for (int i=0; i<NUM_CELLS; i++) {
			Cell cell = new Cell(i,new Vector3D(0,0, 0), this);
			cells.add(cell);
		}
		this.initCellsPositions();
		
	}
	
	/**
	 * Set cell positions and rotations based on their ids
	 */
	public void initCellsPositions() {
		// adjust position & rotation of cells:
		cells.get(0).getPosition().x = 0;
		cells.get(0).getPosition().y = -18;
		
		cells.get(1).getSteeringForce().x = 0.25;
		cells.get(1).getPosition().x = 13;
		cells.get(1).getPosition().y = -13;
		
		cells.get(2).getSteeringForce().x = 0.5;
		cells.get(2).getPosition().x = 18;
		cells.get(2).getPosition().y = 0;
		
		cells.get(3).getSteeringForce().x = 0.75;
		cells.get(3).getPosition().x = 13;
		cells.get(3).getPosition().y = 13;
		
		cells.get(4).getSteeringForce().x = 1;
		cells.get(4).getPosition().x = 0;
		cells.get(4).getPosition().y = 18;
		
		cells.get(5).getSteeringForce().x = -0.75;
		cells.get(5).getPosition().x = -13;
		cells.get(5).getPosition().y = 13;
		
		cells.get(6).getSteeringForce().x = -0.5;
		cells.get(6).getPosition().x = -18;
		cells.get(6).getPosition().y = 0;
		
		cells.get(7).getSteeringForce().x = -0.25;
		cells.get(7).getPosition().x = -13;
		cells.get(7).getPosition().y = -13; 
			
		for (Cell cell: cells) {
			cell.oldWorldPosition = new Vector3D(cell.getWorldPosition());
		}
	}
	
	/**
	 * Main update function. Called each time step
	 */
	public void update() {
		this.oldWorldPosition = new Vector3D (this.getWorldPosition());
		//reset subsumption:
		this.cells.get(8).subsumed = false;
		//reset motor forces:
		for (int i=0; i<4; i++) {
			motorForces[i] = 0;
		}
		
		//---- let all cells update:
		for (Cell cell: cells) {
			cell.update();
		}
		
		//---- make sure motor forces stay in bounds:
		for (int i=0; i<4; i++) {
			if (motorForces[i] > Agent.MAX_MOTOR_FORCE) {
				motorForces[i] = Agent.MAX_MOTOR_FORCE;
			}
			if (motorForces[i] < -Agent.MAX_MOTOR_FORCE) {
				motorForces[i] = -Agent.MAX_MOTOR_FORCE;
			}
		}
		
		//ControlsPanel.getSingletonObject().showOutput(motorForces[MOTOR_LEFT] + "  " + motorForces[MOTOR_RIGHT]);
		if (this.world.getOwner().getTryNo() > -1) {
			//motorForces[Agent.MOTOR_LEFT] = 0.5 *motorForces[Agent.MOTOR_LEFT];
			/*double temp = motorForces[Agent.MOTOR_RIGHT];
			motorForces[Agent.MOTOR_RIGHT] = motorForces[Agent.MOTOR_LEFT];
			motorForces[Agent.MOTOR_LEFT] = temp;*/
			motorForces[Agent.MOTOR_RIGHT] = -motorForces[Agent.MOTOR_RIGHT];
			motorForces[Agent.MOTOR_LEFT] = -motorForces[Agent.MOTOR_LEFT];
			
		}
		
		ControlsPanel.getSingletonObject().showOutput("");
		ControlsPanel.getSingletonObject().showOutput("MOTORS ACTIVATION: L:"+Maths.formatDouble(motorForces[MOTOR_LEFT])+" R:"+Maths.formatDouble(motorForces[MOTOR_RIGHT]));
		//-- save steering force:
		Vector3D oldSteeringForce = new Vector3D(this.getSteeringForce());
		//-- adjust position according to the motor values:
		this.getSteeringForce().y = 0.5*motorForces[Agent.MOTOR_LEFT] + 0.5*motorForces[Agent.MOTOR_RIGHT];
		this.getSteeringForce().x += 0.02*motorForces[Agent.MOTOR_LEFT] - 0.02*motorForces[Agent.MOTOR_RIGHT];
		for (Cell cell: cells) {
			cell.oldWorldPosition = new Vector3D(cell.getWorldPosition());
		}
		//-- move, etc:
		super.update();
		
		//-- adjust position of all cells:
		double angleDiff = (this.getSteeringForce().x - oldSteeringForce.x);
		this.adjustCellsPosition(angleDiff);
		
		//-- decrease food level:
		/*this.foodLevel -= (double)Agent.MAX_FOODLEVEL*(double)0.001;
		if (this.foodLevel < 0) {
			this.foodLevel = 0;
		}
		ControlsPanel.getSingletonObject().showOutput("FOOD: " + Maths.formatDouble(this.foodLevel));
		*/
		
		//register position & speed:
		ReportController.getSingletonObject().registerAgentPosition(this);
		double speed = (Math.hypot(this.getWorldPosition().x - this.oldWorldPosition.x, this.getWorldPosition().y - this.oldWorldPosition.y));
		ReportController.getSingletonObject().registerAgentSpeed(speed);
		
		
		
	}
	
	/**
	 * Adjust local cell positions based on current angle
	 * @param angleDiff
	 */
	private void adjustCellsPosition(double angleDiff) {
		for (Cell cell: cells) {
			Point2D original = new Point2D.Double(cell.getPosition().x, cell.getPosition().y);
			AffineTransform at = AffineTransform.getRotateInstance(Math.toRadians(angleDiff*180), 0, 0);
			Point2D rotated = at.transform(original, null);
			
			cell.getPosition().x = rotated.getX();
			cell.getPosition().y = rotated.getY();
			
		}
	}
	
	//====================================================== ADJUSTMENTS
	/**
	 * Handle event of hitting into an obstacle (actual change of position is done in superclass)
	 */
	protected void handleObjectCollisions(BaseSituatedModel collidedObject)  {
		super.handleObjectCollisions(collidedObject) ;
		
		//if bumped to obstacle, register bump in the report:
		if (collidedObject != null) {
			ReportController.getSingletonObject().registerBumpToObstacle(1, this);
		} else {
			ReportController.getSingletonObject().registerBumpToObstacle(0, this);
		}
	}
	
	//====================================================== MOTORS
	/**
	 * Add to a motor force value
	 */
	public void activateMotor(int motorId, double weight) {
		motorForces[motorId] += Agent.MAX_MOTOR_FORCE * weight;
	}
	
	/**
	 * Adjust position if outside of visible world bounds
	 */
	public void adjustBorderlessWorldCoordins() {
		if (position.x - size.x/4 > Settings.WORLD_W ) {
			position.x = size.x / 4;
		} else if (position.x + size.x/4 < 0 ) {
			position.x = Settings.WORLD_W - size.x / 4;
		}
		
		if (position.y - size.y/4 > Settings.WORLD_H ) {
			position.y = size.y / 4;
		} else if (position.y + size.y /4   < 0 ) {
			position.y = Settings.WORLD_H  - size.y / 4;
		}
	}

	//========================================================== GETTERS / SETTERS
	public ArrayList <Cell> getCells() {
		return cells;
	}
	
	public double[]getMotorForces() {
		return motorForces;
	}
	public String getMotorName(int motorId) {
		String outp = "";
		switch (motorId) {
			case Agent.MOTOR_BOTTOM:
				outp += "B";
				break;
			case Agent.MOTOR_TOP:
				outp += "T";
				break;
			case Agent.MOTOR_LEFT:
				outp += "L";
				break;
			case Agent.MOTOR_RIGHT:
				outp += "R";
				break;
			
		}
		outp += "M";
		return outp;
	}

	public void setFoodLevel(double foodLevel) {
		this.foodLevel = foodLevel;
	}

	public double getFoodLevel() {
		return foodLevel;
	}	
}
