package net.lenkaspace.adaptiveNavigation.model;

import java.util.ArrayList;

import net.lenkaspace.adaptiveNavigation.ControlsPanel;
import net.lenkaspace.adaptiveNavigation.Maths;
import net.lenkaspace.adaptiveNavigation.reports.ReportController;
import net.lenkaspace.adaptiveNavigation.vo.Vector3D;


/** 
 * Cell
 * 	- corresponds to a neuron inside of agent
 * 	- responsible for receiving input and calculating own output to motors
 */

public class Cell extends BaseSituatedModel {
	
	private Agent owner;
	private double activation;
	private double irritation;
	private double output;
	public Vector3D oldWorldPosition;
	public ArrayList<Vector3D> previousPositions;
	private double bias;
	public boolean subsumed;
	
	public static final int BODY_W = 25;
	public static final int BODY_H = 25;
	/* Distance from radial border of cell to radial border of obstacle.
	 * Cell's and obstacle's diameters are added to the range in BaseSituatedModel::getRelativeVectorToObstacle/*
	 */
	public static final int MAX_SENSOR_DISTANCE = (int) Math.round((double)Agent.BODY_W *0.5); 
	public static final double MAX_SENSOR_ANGLE = 0.5; 
	
	public static final int INFORMATION_GAIN_TIME = 150;// 6*Settings.TIME_UNIT_INTERVAL;
	
	
	
	/* STIMULATION MULTIPLIER
	 * strength of current stimulation when added to activation value
	 */
	public static final double STIMULATION_MULTIPLIER_O = 1.8;
	public static final double STIMULATION_MULTIPLIER_F = 0.7;
	
	/* ACTIVATION THETA 
	 *  affects how fast activation grows
	 */
	public static final double ACTIVATION_THETA_O = 20.0;
	public static final double ACTIVATION_THETA_F = 20.0;
	
	/* ACTIVATION THRESHOLD
	 * how sensitive the irritation is towards activity. The greater the number the more activity is allowed before irritation increases
	 */
	public static final double ACTIVATION_THRESHOLD_O = 0.25; 
	public static final double ACTIVATION_THRESHOLD_F = 0.65;
		
	/* IRRITATION THETA 
	 *  affects how fast irritation grows
	 */
	public static final double IRRITATION_THETA_O = 80.0;
	public static final double IRRITATION_THETA_F = 40.0;
	
	/* IRRITATION THRESHOLD
	 *  level of irritation that causes reconfiguration of connections
	 */
	public static final double IRRITATION_THRESHOLD_O = 0.7;
	public static final double IRRITATION_THRESHOLD_F = 0.2; //could vary adaptively with environment
	
	
	/* CONNECTION CHANGE VARIANCE
	 *  maximal variance of Gaussian function that returns new connection value during reconfiguration
	 */
	public static final double CONNECTION_CHANGE_VARIANCE_O = 0.5; // 
	public static final double CONNECTION_CHANGE_VARIANCE_F = 1.5;
	
	/* SUBSUMPTION BORDER
	 * value of activation at which output of an O-neuron subsumes output of F-neuron.
	 * Necessary so that some movement is present when quite far from obstacle but still within range of sensors.
	 * Low values: little movement around walls, high values: restricts obstacle avoidance because of changing F-neuron activation
	 */
	public static final double SUBSUMTION_BORDER = 0.5;
	
	
	private ArrayList<Connection> connectionsToMotors; //dynamic array of connections to motors
	
	public Cell(int id_,Vector3D pos_, Agent owner_)  {
		super(id_, pos_, new Vector3D (Cell.BODY_W,Cell.BODY_H,0), owner_.world);
		owner = owner_;
		this.reset();
		connectionsToMotors = new ArrayList<Connection>();	
		this.initConnections();	
	}
	
	/**
	 * Resets state variables and remembered positions
	 */
	public void reset() {
		activation = 0;
		irritation = 0;
		output = 0;
		bias = 0;
		subsumed = false;
		if (this.id == 8) {
			previousPositions = new ArrayList<Vector3D> ();
			Vector3D worldPosition = this.getWorldPosition();
			for (int i= 0; i< INFORMATION_GAIN_TIME; i++) {
				this.previousPositions.add(new Vector3D((int)worldPosition.x, (int)worldPosition.y, (int)worldPosition.z));
			}
		}
		
		this.oldWorldPosition = new Vector3D(this.getWorldPosition());
	}
	
	/**
	 * Copies attributes from another cell, used during cloning
	 * @param cell
	 */
	public void copyFromCell (Cell cell) {
		bias = cell.bias;
		for (int i=0; i<2; i++) {
			Connection connection = this.connectionsToMotors.get(i);
			connection.copyFromConnection (cell.getConnectionsToMotors().get(i));
		}
	}
	
	/**
	 * Create connections and set their initial values
	 */
	public void initConnections() {
		//----- good connections
		/*if (this.id == 0) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,-1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,-1));
		}  else if (this.id == 1) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,0));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,0));
		} else if (this.id == 2) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,-1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,1));
		} else if (this.id == 3) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,0));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,0));
		} else if (this.id == 4) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,1));
		} else if (this.id == 5) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,0));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,0));
		} else if (this.id == 6) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,-1));
		} else if (this.id == 7) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,0));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,0));
		} */
	
		//---  connections that will force it to bump to obstacles
		/*if (this.id == 0) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,1));
		
		}  else if (this.id == 1) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,1));
		} else if (this.id == 2) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,0.5));
		} else if (this.id == 3) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,-1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,-1));
		} else if (this.id == 4) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,-1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,-1));
		} else if (this.id == 5) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,-1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,-1));
		} else if (this.id == 6) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,0.5));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,1));
		} else if (this.id == 7) {
			connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,1));
			connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,1));
		} */
	
		//--- random connections from range -1;1:
		connectionsToMotors.add(new Connection(Agent.MOTOR_LEFT,Maths.getSingletonObject().getRandomDouble(true)*2-1));
		connectionsToMotors.add(new Connection(Agent.MOTOR_RIGHT,Maths.getSingletonObject().getRandomDouble(true)*2-1));
	}
	
	public Vector3D getWorldPosition() {
		return Vector3D.addVectors(owner.getWorldPosition(), position);
	}
	
	public Vector3D getWorldSteeringForce() {
		return Vector3D.addVectors(owner.getWorldSteeringForce(), steeringForce);
	}
	
	
	
	//====================================== UPDATE
	
	/**
	 * Main update function. Called each time step
	 */
	public void update() {	
		String outputStr = "";
		double noizeInput = 0.1; //possible noize from -noizeInput; noizeInput
		double noizeOutput = 0.1;
		boolean cellNotWorking = false;
		/* ============== Alterations for adaptation experiments */
		if (this.world.getOwner().getTryNo() > 0) {
			if (this.id%2 == 0 && this.id != 8) {
				// cellNotWorking = true;
				// noizeInput = 0.5;
			}
			
		}
		/*==============*/
		boolean showOutput = this.connectionsToMotors.size() > 0;// 
		
		if (!cellNotWorking)
		{
			//------- compute activation based on different kinds of input:
			if (this.id == 8) {
				//---------------- F-neuron:			
				//--- compute flow
				Vector3D flowPosition = this.previousPositions.get(this.previousPositions.size()-100);
				double hyp = Math.hypot(this.getWorldPosition().x - flowPosition.x, this.getWorldPosition().y - flowPosition.y);
				double temp = 100.0*(double)owner.getSpeed()*STIMULATION_MULTIPLIER_F;
				double temp2 = hyp / temp  ;
				double flow = temp2;
				//- add noise:
				flow += (noizeInput*2*Maths.getSingletonObject().getRandomDouble(true)) - noizeInput;
				if (flow < 0) {
					flow = 0;
				} else if (flow > 1) {
					flow = 1;
				} 
				
				//- remove oldest element of memory and add a new one:		
				previousPositions.remove(0);
				Vector3D worldPosition = this.getWorldPosition();
				this.previousPositions.add(new Vector3D((int)worldPosition.x, (int)worldPosition.y, (int)worldPosition.z));	
				
				//- activation:
				this.activation += 1.0/ACTIVATION_THETA_F*(-this.activation + STIMULATION_MULTIPLIER_F*this.computeFastSigmoid(1-flow, 25) + this.bias ); 
				outputStr += "C"+this.id+" F:"+Maths.formatDouble(flow);
			} else {
				//----------------- O-neurons:
				//---- compute distance from nearest object in range:
				double inputNormalised = 0;
				Vector3D v = this.getVectorToClosesObject();
				if (v != null) {
					double distance = Math.abs(v.y);
					//compute stimulation level:
					inputNormalised = 1-(distance/Cell.MAX_SENSOR_DISTANCE);
				}
				//- add noise:
				inputNormalised += (noizeInput*2*Maths.getSingletonObject().getRandomDouble(true)) - noizeInput;
				if (inputNormalised < 0) {
					inputNormalised = 0;
				} else if (inputNormalised > 1) {
					inputNormalised = 1;
				}
				this.activation += 1.0/ACTIVATION_THETA_O*(-this.activation + STIMULATION_MULTIPLIER_O*(inputNormalised) + this.bias ); 
				outputStr += "C"+this.id+" IN:"+Maths.formatDouble(inputNormalised);
			}
	
			//--- clip activation:
			if (this.activation > 1) {
				this.activation = 1;
			} else if (this.activation < 0) {
				this.activation = 0;
			}
			
			//-- subsumption?
			if (this.activation > SUBSUMTION_BORDER && this.id<8 && !cellNotWorking) {
				this.owner.getCells().get(8).subsumed = true;
			}
			
			//---- compute & clip irritation:
			if (this.id == 8) {
				this.irritation += 1.00/IRRITATION_THETA_F * this.computeBrokenLine(ACTIVATION_THRESHOLD_F+ this.bias, this.activation);
			} else {
				this.irritation += 1.00/IRRITATION_THETA_O * this.computeBrokenLine(ACTIVATION_THRESHOLD_O+ this.bias, this.activation);
			}
			if (this.irritation > 1) {
				this.irritation = 1;
			} else if (this.irritation < 0) {
				this.irritation = 0;
			}
			
			//----- adjust connections based on irritation:					
			if (this.id==8) {
				if (this.irritation > Cell.IRRITATION_THRESHOLD_F) {
					this.adjustConnections();
				}
			} else {
				if (this.irritation > Cell.IRRITATION_THRESHOLD_O) {
					this.adjustConnections();
				} 
			}
			
			//--- compute output:
			this.output = this.computeSigmoid(this.activation);		
			if (showOutput) {
				ControlsPanel.getSingletonObject().showOutput(outputStr + " IR:"+Maths.formatDouble(this.irritation)+" A:"+Maths.formatDouble(this.activation)+" O:"+Maths.formatDouble(this.output));
			}
			String motorsStr = "Connect.'s:";
			
			double[] connectionWeights = {0.0,0.0};
			//--- cause all connected motors to fire
			if (!this.subsumed) {
				for (Connection connection : connectionsToMotors) {
					double signalToMotor = this.output*connection.getWeight();
					//-- add noise:
					signalToMotor += (noizeOutput*2*Maths.getSingletonObject().getRandomDouble(true)) - noizeOutput;
					owner.activateMotor(connection.getTargetId(), signalToMotor);
					motorsStr += "  "+owner.getMotorName(connection.getTargetId()) + " Age:"+connection.getAge() +  " W:"+Maths.formatDouble(connection.getWeight());
					connectionWeights[connection.getTargetId()] = connection.getWeight();
					
					//-- add 1 to connection age:
					connection.setAge(connection.getAge() + 1);	
				}
			}
			if (showOutput) {
				ControlsPanel.getSingletonObject().showOutput("  "+ motorsStr);
			}
			
			//------ register reporting values
			ReportController.getSingletonObject().registerWeights(this.id,connectionWeights);
			ReportController.getSingletonObject().registerCellState(this.id, this.activation, this.irritation, this.output);	
			
			
			//decay activation:
			//this.activation -= 0.1;
			if (this.activation < 0) {
				this.activation = 0;
			}
		}		
	}
	
	/**
	 * Set new connection strengths under a certain probability
	 */
	private void adjustConnections() {
		for (Connection connection : this.connectionsToMotors) {
			double probabilityOfChange  = 0.8*(1-(connection.getAge()/Connection.MAX_AGE)) + 0.2; //minimal prob  = 0.2
			if ( Maths.getSingletonObject().getRandomDouble(true) <= probabilityOfChange) {
				double maxVariance = CONNECTION_CHANGE_VARIANCE_O;
				if (this.id == 8) {
					maxVariance = CONNECTION_CHANGE_VARIANCE_F;
				}
				double variance = maxVariance*this.activation;
				//double oldWeight = connection.getWeight();
				connection.setWeight(Maths.getSingletonObject().getRandomGaussian(connection.getWeight(), variance));
			}
			this.irritation = 0;
			
		}
		
	}
	
	
	
	//============================================ MATH FUNCTIONS
	
	/**
	 * computeBrokenLine
	 * 	- line with a break middle, x range 0;1 y range -1;1
	 */
	private double computeBrokenLine(double middle, double in) {
		if (in < middle)  {
	       double temp = 1.0/middle;
	       double result = temp*in - 1;
	       return result;
		} else {
	       double temp = 1.0/(1-middle);
	       double result = temp*in - (temp - 1);
	       return result;
		}
	}
	
	
	/**
	 * computeVshape
	 * 	 - returns val from -1;1 based on a v-shaped graph for inRange 0;1 
	 */
	@SuppressWarnings("unused")
	private double computeVshape(double in) {
		double middle = 0.5;
		if (in >= middle) {
			return in*4 - 3;
		} else {
			return (1-in)*4 - 3;
		}
	}
	
	@SuppressWarnings("unused")
	private double computeVshapeImbalanced(double in) {
		double middle = 0.5;
		if (in >= middle) {
			return in*4 - 3;
		} else {
			return (1-in)*3 - 2.5;
		}
	}
	
	/**
	 * computeVshapeImbalanced2
	 * 	- xRange 0;1, yRange 0;1
	 */
	@SuppressWarnings("unused")
	private double computeVshapeImbalanced2(double in) {
		double middle = 0.4;
		if (in < middle)  {
		       double result = middle - in;
		       return result;
			} else {
		       double temp = 1.0/(1-middle);
		       double result = temp*in - (temp - 1);
		       return result;
			}
	}
	
	/**
	 * computeSigmoid
	 * 	 - returns sigmoid from 0;1 value for inRange 0;1 
	 */
	private double computeSigmoid (double in) {
		double returnVal = (1/ (1 + Math.exp(7 - 13*in)));
		if (in == 0) {
			returnVal = 0;
		} else if (in == 1) {
			returnVal = 1;
		}
		return (returnVal);
	}
	
	/**
	 * computeSigmoid
	 * 	 - returns sigmoid from 0;1 value for inRange 0;1 
	 */
	private double computeFastSigmoid (double in, double slope) {
		double returnVal = (1/ (1 + Math.exp(7 - slope*in)));
		if (in == 0) {
			returnVal = 0;
		} else if (in == 1) {
			returnVal = 1;
		}
		return (returnVal);
	}
	
	/**
	 * Called when getting the nearest object, make sure object is within sensor range
	 */
	protected boolean checkObjectInRange(Vector3D vecToObject) {
		//test if can sense:
		//     orientation     						         &&                      distance in front
		return(Math.abs(vecToObject.x) <= Cell.MAX_SENSOR_ANGLE && vecToObject.y > 0 && vecToObject.y <= Cell.MAX_SENSOR_DISTANCE);
	}
	
	/**
	 * Return true if an object is within range
	 */
	public boolean getIsSensing() {
		BaseSituatedModel object = this.getClosestObject();
		return object!=null;
	}
	
	//============================================ GETTERS / SETTERS
	
	public void setIrritation(double irritation) {
		this.irritation = irritation;
	}

	public double getIrritation() {
		return irritation;
	}
	
	public ArrayList<Connection> getConnectionsToMotors() {
		return connectionsToMotors;
	}
}
